-- Simplistic Object Oriented Programming (SOOP) API Library
-- Version 1.0 Beta
-- Created by: Mario Steele <eumario@trilake.net>
-- Website: http://enchantedblade.trilake.net/
--
-- Filename: soop_class.e
-- Details: This is the core of the Defenition of classes in the
-- SOOP library.  Since SOOP uses Sequences, as it's variables, this
-- already adds to the Natural Object Oriented style of Euphoria, only
-- that it allows for more structure, in a more easy to understand format.

-- /topic Class Definition
-- /info 
--  The core of structuring Data and Routines in SOOP
-- Class definition is the ability to structure Euphoria Data and Routines,
-- into a easy to access method, and improved diversity.  SOOP Classes
-- do not have the same namespace as Euphoria does, so there's no name
-- conflict between Euphoria, and SOOP.
--
-- Some things to note:
--
-- /li Class Block refers to the the definition of a class through the same method as a for loop, while loop, if statement, and such.
-- /li Object Instance refers to a New Instance of a Pre-Defined Class
-- /li When doing a Class Definition, it is not nesscary to declare your method routines as global's, since everything in soop is considered global.
-- /li When using members, Each member is reserved it's own variable space in a object instance.

sequence
	ClassDefs,
	ClassParents,
	ClassMethods,
	ClassMethodRID,
	ClassMembers,
	ClassMemberType,
	ClassConstruct,
	ClassDeconstruct

ClassDefs = {}
ClassParents = {}
ClassMethods = {}
ClassMethodRID = {}
ClassMembers = {}
ClassMemberType = {}
ClassConstruct = {}
ClassDeconstruct = {}

procedure NewClass()
	ClassDefs			= append(ClassDefs,{})
	ClassParents		= append(ClassParents,{})
	ClassMethods		= append(ClassMethods,{})
	ClassMethodRID		= append(ClassMethodRID,{})
	ClassMembers		= append(ClassMembers,{})
	ClassMemberType		= append(ClassMemberType,{})
	ClassConstruct		= append(ClassConstruct,{})
	ClassDeconstruct	= append(ClassDeconstruct,{})
end procedure
	
integer cClass
cClass = 0

--/topic Class Definition
--/proc Class( sfStr ClassName )
--/desc Begins a Class Definition.
--
-- This begins a Class definition, in which you can define various attributes about
-- the class definition, from variables, to routines, that will be specific to that
-- class.  When a new object instance is created, it will receive all the attributes
-- of that class in Memory, seperating it from other objects.  In lamen terms, it's like
-- creating a application within an application.
--
-- Nested Classes are not allowed, but Classes can be parented to multiple diffrent classes.
-- See /Inherit for more details.
global procedure Class(sequence name)
	if not soop_isString(name) then
		soop_error("Class",{name},ERR_STRING)
	end if
	if cClass then	
		soop_error("Class",{name},ERR_INDEF)
	end if
	if find(name,ClassDefs) then
		soop_error("Class",{name},ERR_CLASSDEF)
	end if
	NewClass()
	cClass = length(ClassDefs)
	ClassDefs[cClass] = name
end procedure

-- /topic Class Definition
-- /proc Method( sfStr MethodName, eAtom RoutineID )
-- /desc Attaches a Euphoria Routine to a Class Definition.
--
-- This allows for Routines to be attached to Classes, so that when a object instance is
-- created, the method can easily access the data stored within said object.  This leads
-- to several advantages.
--
-- /li Each class object will have it's own variable space, so that copies of the same class object, does not overwrite each other.
-- /li Methods can easily refer to the object instance from which they are called from, by refering to the read-only 'this' variable.
--
-- Methods can only be defined within a Class Definition block.
global procedure Method(sequence name, atom rid)
	if not soop_isString(name) then
		soop_error("Method",{name,rid},ERR_STRING)
	end if
	if not cClass then
		soop_error("Method",{name,rid},ERR_NINDEF)
	end if
	if find(name,ClassMethods[cClass]) then
		soop_error("Method",{name,rid},{ERR_METHODDEF,name,ClassDefs[cClass]})
	end if
	ClassMethods[cClass] = append(ClassMethods[cClass],name)
	ClassMethodRID[cClass] = append(ClassMethodRID[cClass],rid)
end procedure

-- /topic Class Definition
-- /proc Member( sfStr MemberName, dt DataType )
-- /desc Declares MemberName in the current Class Definition, of DataType.
--
-- Members can become the most powerful part of an Object Oriented Language,
-- they allow for class objects to have their own reserved areas of memory
-- for data to be stored in.  This makes it possible for same like data to
-- be stored in a Euphoria Progam, in a more structured way, without
-- the need to use Constants, and taking away from Euphoria's Namespace issues.
--
-- Members can only be declared within a Class Definition block.
global procedure Member(sequence name, object mtype)
	if not soop_isString(name) then
		soop_error("Member",{name,DataType_GetTypeName(mtype)},ERR_STRING)
	end if
	if not cClass then
		soop_error("Member",{name,DataType_GetTypeName(mtype)},ERR_NINDEF)
	end if
	if find(name,ClassMembers[cClass]) then
		soop_error("Member",{name,DataType_GetTypeName(mtype)},{ERR_MEMBERDEF,name,ClassDefs[cClass]})
	end if
	ClassMembers[cClass] = append(ClassMembers[cClass],name)
	ClassMemberType[cClass] = append(ClassMemberType[cClass],mtype)
end procedure

-- /topic Class Definition
-- /proc Inherit( sfStr ClassName )
-- /desc Allows the current class to Inherit attributes from a parent class.
--
-- Inherit is one of the best things about Object Oriented, it allows you to do
-- sub-classing of a parent class, where you can receive all of the parent's attributes,
-- without having to re-define anything, or re-write code.  Proper polymorphism is done,
-- to allow refrencing to be done with ease of usage.
--
-- Inherit can only be declared within a Class Definition block.
-- When parenting a class, you cannot parent the current class that you are defining,
-- and you cannot parent the same class twice.
global procedure Inherit(sequence name)
	integer class_ix
	if not soop_isString(name) then
		soop_error("Inherit",{name},ERR_STRING)
	end if
	if not cClass then
		soop_error("Inherit",{name},ERR_NINDEF)
	end if
	if find(name,ClassParents[cClass]) then
		soop_error("Inherit",{name},{ERR_AINHERIT,ClassDefs[cClass],name})
	end if
	class_ix = find(name,ClassDefs)
	if class_ix then
		ClassMethods[cClass] = soop_fast_merge_add(ClassMethods[cClass],ClassMethods[class_ix],{-1,name & "::"})
		ClassMethodRID[cClass] = soop_fast_merge(ClassMethodRID[cClass],ClassMethodRID[class_ix])
		ClassMembers[cClass] = soop_fast_merge_add(ClassMembers[cClass],ClassMembers[class_ix],{-1,name & "::"})
		ClassMemberType[cClass] = soop_fast_merge(ClassMemberType[cClass],ClassMemberType[class_ix])
		ClassConstruct[cClass] = soop_fast_merge(ClassConstruct[cClass],ClassConstruct[class_ix])
		ClassDeconstruct[cClass] = soop_fast_merge(ClassDeconstruct[cClass],ClassDeconstruct[class_ix])
		ClassParents[cClass] = append(ClassParents[cClass],name)
	else
		soop_error("Inherit",{name},{ERR_CNDEF,name})
	end if
end procedure

-- /topic Class Definition
-- /proc Constructor( eObj RoutineID )
-- /desc Adds RoutineID to the list of Constructors for a class.
--
-- Constructors allow you a chance to fill information into your Object Instance,
-- and do special things, such as allocate memory, or create Window Handles.
-- Constructors are executed in the order that they are specified, and you can add
-- a single Constructor routine, or add multiple Constructor routines by means of
-- a sequence.
--
-- Constructor can only be defined in a Class Definition block.
global procedure Constructor(object rid)
	if not cClass then
		soop_error("Constructor",{rid},ERR_NINDEF)
	end if
	ClassConstruct[cClass] &= rid
end procedure

-- /topic Class Definition
-- /proc Deconstructor( eObj RoutineID )
-- /desc Adds RoutineID to the list of Deconstructors for a class.
--
-- Like the Constructors, this gives you a chance to free up resources that
-- you may have created during the lifetime of the object.  Deconstructors are
-- executed int he same way as Constructors are, and Deconstructors are executed
-- when a Object Instance is recycled (AKA Destroyed).
--
-- Deconstructors can only be defined in a Class Definition block.
global procedure Deconstructor(object rid)
	if not cClass then
		soop_error("Deconstructor",{rid},ERR_NINDEF)
	end if
	ClassConstruct[cClass] &= rid
end procedure

-- /topic Class Definition
-- /proc EndClass()
-- /desc Terminates the Class Definition Block your currently in.
--
-- This is pretty much self explanitory, it terminates the definition of
-- the current class you are in, and makes it officially available for use
-- by any of the Class routines, or rather, more specifically the new() routine.
global procedure EndClass()
	if not cClass then
		soop_error("EndClass",{},ERR_NINDEF)
	end if
	cClass = 0
end procedure

-- SOOP Functions, for internal libraries.  Should not be used by program, unless you have
-- full knowledge of how these work.

-- /topic Class Low Level Routines
-- /info
--  These are Low Level Routines specifically for use with the Class Definition system.
-- These routines are specifically for use with the Class Definition Core of the
-- SOOP API Library.

-- /topic Class Low Level Routines
-- /func soop_ValidClass( sfStr ClassName )
-- /ret Boolean, True or False
-- /desc Validates weither the Class Name given has been defined or not.
--
-- This can be used in two methods, to figure out if a Class that your
-- library requires, has been defined yet or not, or to ensure that
-- the name isn't already in use.  It's one of the internal function's
-- used by the Object Instance Core library, to confirm weither the
-- class has been defined or not.
global function soop_ValidClass(sequence name)
	if find(name,ClassDefs) then
		return 1
	else
		return 0
	end if
end function

-- /topic Class Low Level Routines
-- /func soop_ClassIX( sfStr ClassName )
-- /ret Atom, Instance of the Class
-- /desc Shorthand way to get the Class' instance in the class table.
--
-- This is used to get the Class Instance in the sequence for Class definitions.
-- This was primarly developed, to remove the need to constantly find the Class
-- instance in other support routines.
global function soop_ClassIX(sequence name)
	return find(name,ClassDefs)
end function

-- /topic Class Low Level Routines
-- /func soop_EnumConstruct( eAtom ClassIX )
-- /ret Sequence containing Routine IDs
-- /desc Retrives the Construct Routines for the Given Class.
--
-- This is the easy way of obtaining the list of Constructors for a class
-- definition.  This was mainly implemented so that I could access the list
-- of constructs, to execute when a new object instance was created.
global function soop_EnumConstruct(atom class_ix)
	return ClassConstruct[class_ix]
end function

-- /topic Class Low Level Routines
-- /func soop_EnumDeconstruct( eAtom ClassIX )
-- /ret Sequence containing Routine IDs
-- /desc Similar to soop_EnumConstruct()
--
-- This does the same thing as soop_EnumConstruct(), only it applies to the
-- deconstruction routines, for when a user destroys a Object instance.
global function soop_EnumDeconstruct(atom class_ix)
	return ClassDeconstruct[class_ix]
end function

-- /topic Class Low Level Routines
-- /func soop_EnumMembers( eAtom ClassIX )
-- /ret Sequence of Strings
-- /desc Returns the list of Members defined on the given Class.
--
-- This returns the names of the members defined on a Class
-- instance.  This was developed, so that it would be easy to
-- grab the Member names, for faster detection in the Object
-- core.
global function soop_EnumMembers(atom class_ix)
	return ClassMembers[class_ix]
end function

-- /topic Class Low Level Routines
-- /func soop_MemberType( eAtom ClassIX, eAtom MemberIX )
-- /ret DataType Constant
-- /desc Returns the Data Type of a Member
--
-- This returns the data type used for a member in the class
-- at the time of it's definition.  It was implemented for
-- easy access to the Type data, without copying the member
-- data to the Object Core Memory.
global function soop_MemberType(atom class_ix, atom m_ix)
	return ClassMemberType[class_ix][m_ix]
end function

-- /topic Class Low Level Routines
-- /func soop_EnumMethods( eAtom ClassIX )
-- /ret Sequence of Strings
-- /desc Returns the Declared Methods on a Class.
--
-- This returns a list of strings that represent the
-- methods defined on a class instance, for faster retrival
-- in the SOOP Object given.
global function soop_EnumMethods(atom class_ix)
	return ClassMethods[class_ix]
end function

-- /topic Class Low Level Routines
-- /func soop_MethodRID( eAtom ClassIX, eAtom MethodIX )
-- /ret Routine ID
-- /desc Returns the Routine ID of a Method
--
-- Returns the Euphoria Routine ID for a Method, for use
-- in method_proc(), method_func(), method_id(), method_call_back().
global function soop_MethodRID(atom class_ix, atom m_ix)
	return ClassMethodRID[class_ix][m_ix]
end function

-- /topic Class Low Level Routines
-- /func soop_EnumClasses()
-- /ret Sequence of Strings
-- /desc Returns the list of currently defined Classes
--
-- This returns the list of currently defined classes in
-- the SOOP Class Enviroment.  It was mainly put in for the
-- SOOP_Debug class in soop_debug.e
global function soop_EnumClasses()
	return ClassDefs
end function

-- /topic Class Low Level Routines
-- /func soop_MemberTypeName( eAtom ClassIX, eAtom MemberIX )
-- /ret String
-- /desc Returns the DataType name for a Member
--
-- Returns the String Identifier for a DataType defined as a
-- member of the Class Instance.  It's mainly a macro, but
-- was implemented for SOOP_Debug class, to make it easier
-- to debug the enviroment with the DataType's actual name.
global function soop_MemberTypeName(atom class_ix, atom m_ix)
	object a
	a = soop_MemberType(class_ix,m_ix)
	if sequence(a) then
		return a
	else
		return DataType_GetTypeName(a)
	end if
end function