-- Simplistic Object Oriented Programming (SOOP) API Library
-- Version 1.0 Beta
-- Created by: Mario Steele <eumario@trilake.net>
-- Website: http://enchantedblade.trilake.net/
--
-- Filename: soop_ll.e
-- Details: These functions here, are the low level stuff used by
-- SOOP.  These functions should /NOT/ be used, unless you have
-- full knowledge of the library's makeup.

global sequence OurError
OurError = {}
sequence MSGS
MSGS = {}

-- /topic SOOP Low Level Routines
-- /info 
--  Low Level routines for use in the SOOP Library, and others.
-- These routines are utilizations of common functions that are required
-- throughout the SOOP Library, and offers good features.  Should you find
-- a need to use them in your own program, feel free to do so, but remember
-- that they do not effect the SOOP Library in any way.

-- /topic SOOP Low Level Routines
-- /func soop_IsString( eSeq data )
-- /ret Boolean True or False
-- /desc Returns True if the Sequence given is a ASCII Printable string
--
-- This function checks the sequence given is a Flat Sequence, that has
-- data stored in it, that is a ASCII Printable string.
global function soop_isString(sequence dat)
	for x = 1 to length(dat) do
		if sequence(dat[x]) then
			return 0
		end if
		if dat[x] < 0 or dat[x] > 255 then
			return 0
		end if
	end for
	return 1
end function

-- /topic SOOP Low Level Routines
-- /func soop_sprints( eObj data )
-- /ret ASCII Printable String
-- /desc Converts a Sequence Data into a ASCII Printable String
--
-- This function is very similar to pretty_print(), however, instead of
-- doing multiple lines, it only does 1 line.  This is commonly used in
-- Error messages, for display the arguments received by a internal function.
global function soop_sprints(object dat)
	sequence ret
	if integer(dat) then
		return sprintf("%d",dat)
	elsif atom(dat) then
		return sprintf("#%d",dat)
	elsif sequence(dat) then
		if soop_isString(dat) then
			return sprintf("\"%s\"",{dat})
		else
			ret = "{"
			for x = 1 to length(dat) do
				ret &= soop_sprints(dat[x])
				if x = length(dat) then
					ret &= "}"
				else
					ret &= ","
				end if
			end for
		end if
	end if
	return ret
end function

-- /topic SOOP Low Level Routines
-- /proc soop_error( sfStr func, eSeq args, eObj msg )
-- /desc Internal SOOP Error Routine.
--
-- This procedure is used internally to display SOOP Errors to the
-- end user.  This program will cause the abort of the program with
-- a 1/0 error, but utilizes the crash_message() routine to display
-- the actual message.
--
-- Example Output: new("MyClassA",{}): Class not defined at this point.
global procedure soop_error(sequence func, sequence args, object msg)
	if integer(msg) then
		msg = MSGS[msg]
	elsif integer(msg[1]) and msg[1] > 0 and msg[1] < length(MSGS)+1 then
		msg = sprintf(MSGS[msg[1]],msg[2..length(msg)])
	elsif sequence(msg[1]) then
		msg = sprintf(msg[1],msg[2..length(msg)])
	end if
	for x = 1 to length(args) do
		if x = 1 then
			args[1] = soop_sprints(args[x])
		else
			args[1] &= soop_sprints(args[x])
		end if
		if x != length(args) then
			args[1] &= ","
		end if
	end for
	args = args[1]
	OurError = sprintf("%s(%s): %s",{func,args,msg})
	machine_proc(37,OurError)
	? 1/0
end procedure

-- /topic SOOP Low Level Routines
-- /func soop_adderror( eSeq msg )
-- /ret Identifier pointer to message
--
-- This function is used to decleare a new Error Message, oftenly used with
-- soop_error() as a way to display error messages for easier display of messages.
-- This should only be used if you know what youa re doing.
global function soop_adderror(sequence msg)
	MSGS &={ msg }
	return length(MSGS)
end function

-- /topic SOOP Low Level Routines
-- /func soop_reverse( eSeq dat )
-- /ret Returns the sequence in reversed order.
--
-- This is just like the reverse() defined in misc.e, only supplied here
-- to keep from including extra libraries, that are not nesscary.
global function soop_reverse(sequence dat)
	integer lower, n1, n2
	sequence t
	
	n1 = length(dat)
	n2 = floor(n1/2)+1
	lower = 1
	t = repeat(0,n1)
	for upper = n1 to n2 by -1 do
		t[lower] = dat[upper]
		t[upper] = dat[lower]
		lower += 1
	end for
	return t
end function

-- /topic SOOP Low Level Routines
-- /func soop_fast_merge( eSeq Source, eSeq New)
-- /ret Merged sequence of Source and New
--
-- This function merges two sequences, or sequence parts, into 1 sequence.
-- It uses a fast merging method as displayed in reverse().
global function soop_fast_merge(sequence s, sequence n)
	integer lower, n1, n2, o
	sequence t
	
	n1 = length(n)
	n2 = floor(n1/2)+1
	t = s & repeat(0,n1)
	o = length(s)
	lower = 1
	for upper = n1 to n2 by -1 do
		t[upper+o] = n[upper]
		t[lower+o] = n[lower]
		lower += 1
	end for
	return t
end function

-- /topic SOOP Low Level Routines
-- /func soop_fast_merge_add( eSeq Source, eSeq New, eSeq Extra)
-- /ret Merged sequence of Source and New, appending the Extra to the New
--
-- Much like soop_fast_merge(), this function uses the same method as
-- reverse() to merge two sequences into 1, or two sequence parts into 1.
-- The diffrence is, though that all New objects will get the Extra data
-- added to the New Object.
--
-- Extra can be a 2 part sequence, where the first part determins where
-- in the New sequence, the data should be added.
--
-- Example:
-- /code
-- sequence s
-- s = soop_fast_merge_add({"Apple","Orange"},{"Lemon","Lime"},{0,"Sour "})
-- -- s is now {"Apple","Orange","Sour Lemon","Sour Lime"}
-- s = soop_fast_merge_add({"Lemon","Lime"},{"Watermelon","Apple"},{-1," Sweet"})
-- -- s is now {"Lemon","Lime","Watermelon Sweet","Apple Sweet"}
-- /endcode
global function soop_fast_merge_add(sequence s, sequence n, sequence dat)
	integer lower, n1, n2, o
	sequence t
	n1 = length(n)
	n2 = floor(n1/2)+1
	t = s & repeat(0,n1)
	o = length(s)
	lower = 1
	if dat[1] = 0 then
		for upper = n1 to n2 by -1 do
			t[upper+o] = n[upper] & dat[2]
			t[lower+o] = n[lower] & dat[2]
			lower += 1
		end for
	elsif dat[1] = -1 then
		for upper = n1 to n2 by -1 do
			t[upper+o] = dat[2] & n[upper]
			t[lower+o] = dat[2] & n[lower]
			lower += 1
		end for
	else
		for upper = n1 to n2 by -1 do
			t[upper+o] = n[upper][1..dat[1]] & dat[2] & n[upper][dat[1]..length(n[upper])]
			t[lower+o] = n[lower][1..dat[1]] & dat[2] & n[lower][dat[1]..length(n[lower])]
			lower += 1
		end for
	end if
	return t
end function

-- /topic SOOP Low Level Routines
-- /func soop_iff( eInt Expresion, eObj True, eObj False )
-- /ret Returns based apon Expression.
--
-- This function is basically like an if statement, in how it evaluates the
-- expression, then returns what is in the True variable, if the expression is
-- true, or returns the False variable, if the expression is False.
--
-- Example:
-- /code
-- puts(1,soop_iff(1+1=2,"True","False"))
-- -- Prints onto the screen "True"
-- puts(1,soop_iff(1+2=2,"True","False"))
-- -- Prints onto the screen "False"
-- /endcode
global function soop_iff(integer tf, object t, object f)
	if tf then
		return t
	else
		return f
	end if
end function