-- Simplistic Object Oriented Programming (SOOP) API Library
-- Version 1.0 Beta
-- Created by: Mario Steele <eumario@trilake.net>
-- Website: http://enchantedblade.trilake.net/
--
-- Filename: soop_object.e
-- Details: This controls the Object Instances of each class.  This is
-- really the core of the SOOP library, as this controls the actual
-- instance creation.

-- /topic Object Manipulation
-- /info
--  Routines for Creating, Manipulating and Destroying Class Objects.
-- This is the core of the SOOP API Library.  These functions are what
-- allow you to do most of the work that is needed in a Object Oriented
-- library.

sequence
	ObjectClass,	-- Class Name, for faster error reporting
	ObjectClassIX,	-- Class IX, for Faster reading
	ObjectMembers,	-- Member Name, for faster error reporting
	ObjectVTBL,		-- Member Data Table
	ObjectMethods,	-- Method Names, for faster access
	ObjectLAST,		-- Last Value retrived from oget()
	
	MethodID		-- For method_id() calling

ObjectClass = {}
ObjectClassIX = {}
ObjectMembers = {}
ObjectVTBL = {}
ObjectMethods = {}
ObjectLAST = {}
MethodID = {}

atom thisVar
thisVar = 0

constant OBJECT_MEMBERS = 1,
         OBJECT_METHODS = 0

-- /topic Polymorphism
-- /info
--  Information for how SOOP handles Polymorphism
--
-- The SOOP library allows for Polymorphism to be used, as a way to
-- allow Overloading of Members, and Methods.  Overloading can be defined
-- as in a object oriented enviroment, as defining multiple members of the
-- like name, or multiple methods of like name, while still allowing for
-- the original to be there.  This is especially useful when you inherit
-- other classes in the class your working with.
--
-- There's two forms of Polymorphism.  The first form, involves Inheritance.
-- This form of polymorphism uses the token identifier '::', as a way to define
-- the structure, or if you would, path to the object your looking for.
--
-- For example, should you define a class by "MyClassA", and define a member
-- "Apple" in it, then create a New class called "MyClassB", which Inherits from
-- "MyClassA", when the inheritance occurs, The member "Apple" is decorated as
-- "MyClassA::Apple".  This is done, so should you decide to declare another
-- member in MyClassB, as "Apple", you can do it, without getting Name confliction.
--
-- Should you want to refrence the Apple member in MyClassA, you would use
-- "MyClassA::Apple".  If you don't have a member named "Apple" in MyClassB,
-- you can Refer to "Apple", and get the value of "Apple" in MyClassA's definition.
-- However, if you have a member "Apple" declared in MyClassB, you would get
-- MyClassB Apple's value.
--
-- This form of polymorphism also applies to methods.
--
-- The second form of Polymorphism that is used in SOOP, is the form in which
-- member names are used, as a fast, and easier way to follow, to call methods,
-- or get member values from classes that are defined as members of the class.
--
-- Using the Above Example, of MyClassA being defined, and the Member Apple being
-- defined in MyClassA, but changing it, to where instead of inheriting MyClassA
-- in MyClassB, you make it a Member called ClassB.  You can retrive the value of
-- Apple from ClassB, by using "ClassB->Apple".  This makes it so that you do
-- not have to use multiple oget()'s, or oset()'s on classes that you don't nesscarly
-- need to.
--
-- This may not follow normal Object Oriented Polymorphism, but is viewed as
-- the way to do it in the SOOP Library, and should be called Pathing Polymorphism.

-- SOOP Support Functions
function soop_GetEndPoly(sequence name)
	if match("::",name) then
		for End = length(name) to 1 by -1 do
			if End = 1 then
				return name
			end if
			if name[End] = ':' and
				name[End-1] = ':' then
				return name[End+1..length(name)]
			end if
		end for
	else
		return name
	end if
end function

function soop_FindPoly(atom ix, sequence name, integer which)
	if which = OBJECT_MEMBERS then
		for x = 1 to length(ObjectMembers[ix]) do
			if equal(name,ObjectMembers[ix][x]) then
				return x
			elsif equal(name,soop_GetEndPoly(ObjectMembers[ix][x])) then
				return x
			end if
		end for
	elsif which = OBJECT_METHODS then
		for x = 1 to length(ObjectMethods[ix]) do
			if equal(name,ObjectMethods[ix][x]) then
				return x
			elsif equal(name,soop_GetEndPoly(ObjectMethods[ix][x])) then
				return x
			end if
		end for
	end if
	return 0
end function

function soop_ResolvePoly(atom ix, sequence name)
	sequence temp, tempa, tempb, tempc
	integer lower, n, n1, found
	while match("->",name) do
		found = 0
		temp = name[1..match("->",name)-1]
		tempa = ObjectMembers[ix]
		lower = 1
		n = length(tempa)
		n1 = floor(n/2)+1
		for upper = n to n1 by -1 do
			tempb = soop_GetEndPoly(tempa[upper])
			tempc = soop_GetEndPoly(tempa[lower])
			if equal(temp,tempb) then
				ix = ObjectVTBL[ix][upper]
				name = name[match("->",name)+2..length(name)]
				found = 1
				exit
			elsif equal(temp,tempc) then
				ix = ObjectVTBL[ix][lower]
				name = name[match("->",name)+2..length(name)]
				found = 1
				exit
			end if
			lower += 1
		end for
		if not found then
			return {-1,ix,name,temp}
		end if
		if ix < 0 or ix > length(ObjectClass) then
			return {0,ix,name,temp}
		end if
	end while
	return {ix,name}
end function

procedure soop_NewInstance()
	ObjectClass = append(ObjectClass,-1)
	ObjectClassIX = append(ObjectClassIX,-1)
	ObjectMembers = append(ObjectMembers,-1)
	ObjectVTBL = append(ObjectVTBL,-1)
	ObjectMethods = append(ObjectMethods,-1)
	ObjectLAST = append(ObjectLAST,-1)
end procedure

function soop_GetIX()
	return find(-1,ObjectClass)
end function

procedure soop_FreeInstance(atom ix)
	ObjectClass[ix] = -1
	ObjectClassIX[ix] = -1
	ObjectMembers[ix] = -1
	ObjectVTBL[ix] = -1
	ObjectMethods[ix] = -1
	ObjectLAST[ix] = -1
end procedure

procedure soop_FreeMethodID(atom ix)
	integer lower, n, n1
	
	n = length(MethodID)
	n1 = floor(n/2)+1
	lower = 1
	for upper = n to n1 by -1 do
		if MethodID[upper][1] = ix then
			MethodID[upper] = -1
		end if
		if MethodID[lower][1] = ix then
			MethodID[lower] = -1
		end if
		lower += 1
	end for
end procedure

-- /topic Object Manipulation
-- /func ValidObject( sObj Instance )
-- /ret Boolean True or False
-- /desc This function determins weither or not a Object is Valid.
--
-- It checks the Object memory space, to ensure that the object is
-- a valid object, and returns true if it exists, or false if it doesn't
-- exist.  There's two ways that an Object may not exist.  If the
-- object has been removed from memory via the recycle procedure,
-- or has not been created yet.
global function ValidObject(atom ix)
	if ix > 0 and ix < length(ObjectClass)+1 then
		if sequence(ObjectClass[ix]) then
			return 1
		end if
	end if
	return 0
end function

-- /topic Object Manipulation
-- /func ClassType( sObj Instance )
-- /ret String Class Name
-- /desc Returns the Class Name of the Object Instance.
--
-- This returns the Class Name used to create the object instance
-- given.  This makes it easier to ensure that you are dealing with
-- the right class or not.
global function ClassType(atom ix)
	if ValidObject(ix) then
		return ObjectClass[ix]
	else
		soop_error("ClassType",{ix},ERR_OBJ_INVOBJ)
	end if
end function

-- Core Memory Space creation/destruction of data.

-- /topic Object Manipulation
-- /func new( sfStr ClassName, eSeq CreateArguments )
-- /ret Returns an Object Instance
-- /desc Creates a New object given by Class, and returns it's instance if successfully created.
--
-- This function creates a new object instance of the class definition given by ClassName.
-- CreateArguments are the arguments that are passed off to your Construct routine, which is
-- executed after all the base data has been stored, allowing you to modify information as
-- needed.
global function new(sequence class, sequence dat)
	atom ix, nix, class_ix, oldThis
	sequence temp
	object mType
	if not soop_isString(class) then
		for x = 1 to length(class) do
			if not soop_isString(class[x]) then
				soop_error("new",{class,dat},ERR_STRING)
			end if
		end for
		if length(class) != length(dat) then
			soop_error("new",{class,dat},ERR_OBJ_ARGLENGTH)
		end if
		temp = {}
		for x = 1 to length(class) do
			temp &= new(class[x],dat[x])
		end for
		return temp
	end if
	ix = soop_GetIX()
	if ix = 0 then
		soop_NewInstance()
		ix = soop_GetIX()
	end if
	if not soop_ValidClass(class) then
		soop_error("new",{class,dat},ERR_OBJ_NOCLASS)
	end if
	ObjectClass[ix] = class
	class_ix = soop_ClassIX(class)
	ObjectClassIX[ix] = class_ix
	ObjectMethods[ix] = soop_EnumMethods(class_ix)
	temp = soop_EnumMembers(class_ix)
	ObjectMembers[ix] = temp
	ObjectVTBL[ix] = repeat(-1,length(temp))
	for x = 1 to length(temp) do
		mType = soop_MemberType(class_ix,x)
		if sequence(mType) then
			nix = new(mType,{dat})
			ObjectVTBL[ix][x] = nix
		else
			ObjectVTBL[ix][x] = DataType_InitVar(mType)
		end if
	end for
	temp = soop_EnumConstruct(class_ix)
	oldThis = thisVar
	thisVar = ix
	for x = 1 to length(temp) do
		call_proc(temp[x],{dat})
	end for
	thisVar = oldThis
	return ix
end function

-- /topic Object Manipulation
-- /proc recycle( sObj Instance )
-- /desc Destroys a Object Instance given.
--
-- This destroys the object instance given, and returns the memory used
-- by the object to the library for future object instance creation.
-- All Deconstructors that are attached to a object are executed at
-- this time, and any method_id()'s used by this object also become
-- invalid during this time.
global procedure recycle(atom ix)
	sequence destroys
	object mType
	atom oldThis
	if not ValidObject(ix) then
		soop_error("recycle",{ix},ERR_OBJ_INVOBJ)
	end if
	for x = 1 to length(ObjectMembers[ix]) do
		mType = soop_MemberType(ObjectClassIX[ix],ix)
		if sequence(mType) then
			recycle(ObjectVTBL[ix][x])
		else
			DataType_FreeVar(mType,ObjectVTBL[ix][x])
		end if
	end for
	destroys = soop_EnumDeconstruct(ObjectClassIX[ix])
	oldThis = thisVar
	thisVar = ix
	for x = 1 to length(destroys) do
		call_proc(destroys[x],{})
	end for
	thisVar = oldThis
	soop_FreeMethodID(ix)
	soop_FreeInstance(ix)
end procedure

-- Core Memory I/O for Instance data.

-- /topic Object Manipulation
-- /func oget( sObj Instance, eSeq MemberName )
-- /ret Data stored in Member
-- /desc This returns the data stored in a Member of Object.
--
-- This allows for same name variables to be declared, and seperated by
-- the instance object.  It allows for data to be retrived of any type,
-- and dependant apon the Data Type given, you can also retrive the
-- pointer used by the Member.
--
-- Example:
-- /code
-- Class("MyClassA")
--    Member("Var1",STRING)
-- EndClass()
-- constant MyObj = new("MyClassA")
-- oset(MyObj,"Var1","Hello World")
-- c_proc(xSetWindowTitle,{hWnd,oget(MyObj,{"Hello World"})})
-- -- The title of hWnd is now "Hello World", but is set via the Pointer returned by oget()
-- /endcode
--
-- Multiple refrences can also be passed to name, so you could do something
-- like this:
-- /code
-- Class("Score")
--    Member("Score",INTEGER)
--    Member("Name",STRING)
--    procedure Score_construct(sequence dat)
--      oset(this,"Name",dat[1])
--      oset(this,"Score",dat[2])
--    end procedure
--    Constructor(routine_id("Score_construct"))
-- EndClass()
-- constant Player = new("Score",{"Mario",32})
-- printf(1,"%s's score is %d",oget(Player,{"Name","Score"}))
-- -- Displays: "Mario's score is 32"
-- /endcode
global function oget(atom ix, sequence name)
	integer GetPointer
	sequence temp
	atom m_ix
	object mType
	if ix = this then
		ix = thisVar
	end if
	if not ValidObject(ix) then
		soop_error("oget",{ix,name},ERR_OBJ_INVOBJ)
	end if
	GetPointer = 0
	if not soop_isString(name) then
		if length(name) = 1 then
			if soop_isString(name[1]) then
				GetPointer = 1
				name = name[1]
			else
				soop_error("oget",{ix,name},ERR_STRING)
			end if
		else
			temp = {}
			for x = 1 to length(name) do
				temp &= oget(ix,name[x])
			end for
			ObjectLAST[ix] = temp
			return temp
		end if
	end if
	temp = soop_ResolvePoly(ix,name)
	if temp[1] = -1 then
		soop_error("oget",{ix,name},{ERR_OBJ_MEM404,temp[4],temp[2]})
	elsif temp[1] = 0 then
		soop_error("oget",{ix,name},ERR_OBJ_INVOBJ)
	else
		ix = temp[1]
		name = temp[2]
	end if
	m_ix = soop_FindPoly(ix,name,OBJECT_MEMBERS)
	if m_ix then
		mType = soop_MemberType(ObjectClassIX[ix],m_ix)
		if sequence(mType) then
			ObjectLAST[ix] = ObjectVTBL[ix][m_ix]
			return ObjectLAST[ix]
		else
			if GetPointer then
				if DataType_AllowPointer(mType) then
					ObjectLAST[ix] = ObjectVTBL[ix][m_ix]
					return ObjectLAST[ix]
				else
					soop_error("oget",{ix,name},ERR_OBJ_DTNOPOINT)
				end if
			else
				ObjectLAST[ix] = DataType_Retrieve(mType,ObjectVTBL[ix][m_ix])
				return ObjectLAST[ix]
			end if
		end if
	else
		soop_error("oget",{ix,name},{ERR_OBJ_MEM404,name,ix})
	end if
end function

-- /topic Object Manipulation
-- /func olast( sObj Instance )
-- /ret Returns the data in a Member
-- /desc Returns the last data retrived with oget() on the object instance.
--
-- This function maybe useful, if you use the same member name, in a if elsif end if statement.
-- It returns the last data retrived with oget().
--
-- Example:
-- /code
-- Class("ClassA")
--   Member("Platform",INTEGER)
--   procedure Constructor_ClassA(sequence dat)
--     oset(this,"Platform",platform())
--   end procedure
--   Constructor(routine_id("Constructor_ClassA"))
-- EndClass()
-- constant Platform = new("ClassA",{})
-- if oget(Platform,"Platform",{}) then end if
-- if olast(Platform) = DOS32 then
--   puts(1,"You are running on DOS32 Euphoria.")
-- elsif olast(Platform) = WIN32 then
--   puts(1,"You are running on Win32 Euphoria.")
-- elsif olast(Platform) = LINUX then
--   puts(1,"You are running on Linux/FreeBSD Euphoria.")
-- end if
-- /endcode
global function olast(atom ix)
	if ix = this then
		ix = thisVar
	end if
	if ValidObject(ix) then
		return ObjectLAST[ix]
	else
		soop_error("oget",{ix},ERR_OBJ_INVOBJ)
	end if
end function

-- /topic Object Manipulation
-- /proc oset( sObj Instance, eSeq MemberName, eObj Value )
-- /desc Stores data into a Member on a Object Instance.
--
-- This procedure allows you to store data of Value in a member given by
-- MemberName, on an object given by Instance.  Multiple refrences and
-- values can be given.
--
-- Example:
-- /code
-- Class("MyClassA")
--   Member("Var1",INTEGER)
--   Member("Var2",ATOM)
-- EndClass()
-- constant Obj = new("MyClassA",{})
-- oset(Obj,{"Var1","Var2"},{32,#FFFFFF})
-- /endcode
global procedure oset(atom ix, sequence name, object val)
	integer SetPointer
	sequence temp
	atom m_ix
	object mType
	if ix = this then
		ix = thisVar
	end if
	if not ValidObject(ix) then
		soop_error("oset",{ix,name,val},ERR_OBJ_INVOBJ)
	end if
	SetPointer = 0
	if not soop_isString(name) then
		if length(name) = 1 then
			if soop_isString(name[1]) then
				SetPointer = 1
				name = name[1]
			else
				soop_error("oset",{ix,name,val},ERR_STRING)
			end if
		else
			if length(name) != length(val) then
				soop_error("oset",{ix,name,val},ERR_OBJ_ARGLENGTH)
			end if
			for x = 1 to length(name) do
				oset(ix,name[x],val[x])
			end for
			return
		end if
	end if
	temp = soop_ResolvePoly(ix,name)
	if temp[1] = -1 then
		soop_error("oset",{ix,name,val},{ERR_OBJ_MEM404,temp[4],temp[2]})
	elsif temp[1] = 0 then
		soop_error("oset",{ix,name,val},ERR_OBJ_INVOBJ)
	else
		ix = temp[1]
		name = temp[2]
	end if
	m_ix = soop_FindPoly(ix,name,OBJECT_MEMBERS)
	if m_ix then
		mType = soop_MemberType(ObjectClassIX[ix],m_ix)
		if sequence(mType) then
			if sequence(val) then
				val = val[1]
				SetPointer = 1
			end if
			if ValidObject(val) then
				if SetPointer then
					recycle(ObjectVTBL[ix][m_ix])
				end if
				ObjectVTBL[ix][m_ix] = val
			else
				soop_error("oset",{ix,name,val},ERR_OBJ_INVOBJ)
			end if
		else
			if SetPointer then
				if DataType_AllowPointer(mType) then
					ObjectVTBL[ix][m_ix] = val
				else
					soop_error("oset",{ix,name,val},ERR_OBJ_DTNOPOINT2)
				end if
			else
				ObjectVTBL[ix][m_ix] = DataType_Store(mType,ObjectVTBL[ix][m_ix],val)
			end if
		end if
	else
		soop_error("oset",{ix,name,val},{ERR_OBJ_MEM404,name,ix})
	end if
end procedure

function soop_ReplaceOIData(atom obj_ix)
	sequence temp
	object Type
	temp = {ObjectClass[obj_ix],ObjectVTBL[obj_ix]}
	for x = 1 to length(temp[2]) do
		Type = soop_MemberType(ObjectClassIX[obj_ix],x)
		if sequence(Type) then
			temp[2][x] = soop_ReplaceOIData(temp[2][x])
		else
			temp[2][x] = {Type,DataType_Retrieve(Type,temp[2][x])}
		end if
	end for
	return temp
end function

procedure soop_RestoreOIData(sequence entry, object parms, atom ix, sequence obj)
	object Type
	for x = 1 to length(obj) do
		if x > length(ObjectVTBL[ix]) then
			soop_error(entry,parms,"Class definition for object has been changed!")
		end if
		Type = soop_MemberType(ObjectClassIX[ix],x)
		if sequence(obj[x][1]) then
			if not equal(obj[x][1],Type) then
				soop_error(entry,parms,"Class definition for object member has been changed!")
			end if
			soop_RestoreOIData(entry,parms,ObjectVTBL[ix][x],obj[x][2])
		else
			if obj[x][1] != Type then
				soop_error(entry,parms,"Class definition for object has been changed!")
			end if
			ObjectVTBL[ix][x] = DataType_Store(Type,ObjectVTBL[ix][x],obj[x][2])
		end if
	end for
end procedure

function soop_RestoreClassTree(sequence obj)
	sequence class, temp
	atom ix, ix2, class_ix
	class = obj[1]
	obj = obj[2]
	if not soop_ValidClass(class) then
		return 0
	end if
	ix = soop_GetIX()
	if ix = 0 then
		soop_NewInstance()
		ix = soop_GetIX()
	end if
	ObjectClass[ix] = class
	class_ix = soop_ClassIX(class)
	ObjectClassIX[ix] = class_ix
	ObjectMethods[ix] = soop_EnumMethods(class_ix)
	temp = soop_EnumMembers(class_ix)
	ObjectMembers[ix] = temp
	ObjectVTBL[ix] = repeat(-1,length(temp))
	for x = 1 to length(obj) do
		if sequence(obj[x][1]) then
			ix2 = soop_RestoreClassTree(obj[x])
			if ix2 = 0 then
				return 0
			end if
			ObjectVTBL[ix][x] = ix2
		else
			ObjectVTBL[ix][x] = DataType_InitVar(soop_MemberType(class_ix,x))
		end if
	end for
	return ix
end function

-- Core instance saving/reading functions

-- /topic Object Manipulation
-- /proc owrite( eAtom FileHandle, sObj Instance )
-- /desc Writes an Object Instance to File.
--
-- This routine allows you to write a Object Instance's data to disk
-- for later reading back into the program, or by another program.  There's
-- several thing to notes however, that when writting a objects data to disk
-- Temporary data, such as Window Handles, Allocated Memory Pointers, and
-- File handles will not be valid in another program, or not valid when you run
-- the program again.
--
-- Pointers from the Data Type system will never be written to file, the actual
-- data stored in them will be though.
--
-- Data Corruption can still happen.  The library tries to detect as much as
-- possible, but it's not a sure thing.  It's reliable, but there's still a
-- a chance that the data will get corrupt.
--
-- Classes defined as Members will get their data structure also written to disk
-- when the library writes the instance in question to disk.
global procedure owrite(atom fh, atom ix)
	sequence obj
	if ix = this then
		ix = thisVar
	end if
	if not ValidObject(ix) then
		soop_error("owrite",{fh,ix},ERR_OBJ_INVOBJ)
	end if
	obj = soop_ReplaceOIData(ix)
	PutObject(fh,obj)
end procedure

-- /topic Object Manipulation
-- /func oread( eAtom FileHandle, sObj Instance )
-- /ret Object Instance
-- /desc Reads a Object stored in a file, back into a Instance in Memory.
--
-- This will read a Object Instance from a file, given by FileHandle, and
-- store it back into the Object Memory, so that you can use it.  Instance
-- can be a valid Object, as long as it is the same class that is being
-- read in, or you can supply 0 to indicate that a new instance to be
-- created from the object data.
--
-- Constructors are not executed if the oread() function must create
-- a new instance.
--
-- Classes that are members of the instance being read in, will get
-- the data restored, or recreated, when the data is being read in.
global function oread(atom fh, atom ix)
	sequence class
	object obj
	if ix = this then
		ix = thisVar
	end if
	obj = GetObject(fh)
	if not sequence(obj) then
		soop_error("oread",{fh,soop_iff(ValidObject(ix),ix,"<New Instance>")},"Unable to read Object data from file given.")
	end if
	if not ValidObject(ix) then
		ix = soop_RestoreClassTree(obj)
		if ix = 0 then
			soop_error("oread",{fh,"<New Instance>"},"Unable to restore Object Data to Structure, due to an Invalid Class definition.")
		end if
	end if
	class = obj[1]
	obj = obj[2]
	if not equal(ObjectClass[ix],class) then
		soop_error("oread",{fh,ix},"Object instance class does not match object being read in!")
	end if
	soop_RestoreOIData("oread",{fh,ix},ix,obj)
	return ix
end function

-- Core instance memory writting/reading functions

-- /topic Object Manipulation
-- /func oallocate( sObj Instance )
-- /ret Memory Location
-- /desc Stores the Object Instance into Memory
--
-- This converts the Object Instance data, into a ASCII String, and stores
-- it into Memory, which can be used to send across the internet, or
-- through other measures, in which to send or receive data.  The same rules
-- for this, applies as they do with owrite().
global function oallocate(atom ix)
	atom mem
	sequence obj
	if ix = this then
		ix = thisVar
	end if
	if not ValidObject(ix) then
		soop_error("oallocate",{ix},ERR_OBJ_INVOBJ)
	end if
	obj = soop_ReplaceOIData(ix)
	mem = AllocateObject(obj)
	return mem
end function

-- /topic Object Manipulation
-- /func opeek( eAtom MemoryAddress, sObj Instance)
-- /ret Object Instance
-- /desc Reads a Object Instance Data from a Memory Location
--
-- This converst he Object Instance data in an ASCII String memory location,
-- back into a Object Instance.  This function follows the same rules as
-- oread() does.
global function opeek(atom mem, atom ix)
	object obj
	sequence class
	obj = PeekObject(mem)
	if not sequence(obj) then
		soop_error("opeek",{mem,ix},"Unable to peek Object data from memory location given.")
	end if
	if not ValidObject(ix) then
		ix = soop_RestoreClassTree(obj)
		if ix = 0 then
			soop_error("opeek",{mem,"<New Instance>"},"Unable to restore Object Data to Structure, due to an Invalid Class definition.")
		end if
	end if
	class = obj[1]
	obj = obj[2]
	if not equal(ObjectClass[ix],class) then
		soop_error("opeek",{mem,ix},"Object instance class does not match object being read in!")
	end if
	soop_RestoreOIData("opeek",{mem,ix},ix,obj)
	return ix
end function

-- Core instance sequence writting/reading functions

-- /topic Object Manipulation
-- /func oencode( sObj Instance )
-- /ret Sequence Flat String
-- /desc Converts an Object Instance into an ASCII Encoded Sequence Flat String.
--
-- This allows Object Instance data to be returned as an ASCII Encoded String, to
-- the caller.  This can be used to manipulate the data,or do various other things.
-- This function operates in the same method and rules as owrite() and oallocate().
global function oencode(atom ix)
	sequence obj
	if ix = this then
		ix = thisVar
	end if
	if not ValidObject(ix) then
		soop_error("oencode",{ix},ERR_OBJ_INVOBJ)
	end if
	obj = soop_ReplaceOIData(ix)
	return EncodeObject(obj)
end function

-- /topic Object Manipulation
-- /func odecode( eSeq dat, sObj Instance )
-- /ret Object Instance
-- /desc Converts an Object Instance back from a ASCII Encoded Sequences Flat String.
--
-- This allows Object Instance data to be put back into an Object memory location,
-- in the same form that it was written into.  This function follows the same rules
-- as oread() and opeek().
global function odecode(sequence dat, atom ix)
	object obj
	sequence class
	obj = DecodeObject(dat)
	if not sequence(obj) then
		soop_error("odecode",{"<Object Data>",ix},"Unable to peek Object data from memory location given.")
	end if
	if not ValidObject(ix) then
		ix = soop_RestoreClassTree(obj)
		if ix = 0 then
			soop_error("odecode",{"<Object Data>","<New Instance>"},"Unable to restore Object Data to Structure, due to an Invalid Class definition.")
		end if
	end if
	class = obj[1]
	obj = obj[2]
	if not equal(ObjectClass[ix],class) then
		soop_error("odecode",{"<Object Data>",ix},"Object instance class does not match object being read in!")
	end if
	soop_RestoreOIData("odecode",{"<Object Data>",ix},ix,obj)
	return ix
end function

-- Core Execution of methods

-- /topic Object Manipulation
-- /proc method_proc( sObj Instance, eSeq MethodName, eSeq Arguments)
-- /desc Calls a Method on a Object Instance.
--
-- This routine allows you to call a Method defined as an Euphoria
-- procedure.  It follows the same rules as call_proc() does, however
-- instead of using a routine_id() as an argument to this, you use the
-- method's name, as the caller.  Multiple Methods can be called with
-- this as well.
--
-- Example:
-- /code
-- Class("MyClassA")
--   procedure doSomething()
--      puts(1,"Hello World!\n")
--   end procedure
--   Method("display1",routine_id("doSomething"))
--   procedure doSomething2()
--      puts(1,"How are you today?")
--   end procedure
--   Method("display2",routine_id("doSomething2"))
-- EndClass()
-- constant aba = new("MyClassA",{})
-- method_proc(aba,{"display1","display2"},{{},{}})
-- -- displays:
-- -- Hello World!
-- -- How are you today?
-- /endcode
global procedure method_proc(atom ix, sequence name, sequence args)
	sequence temp
	atom oldThis, m_ix
	if ix = this then
		ix = thisVar
	end if
	if not soop_isString(name) then
		if length(name) = 1 then
			soop_error("method_proc",{ix,name,args},ERR_STRING)
		end if
		if length(name) != length(args) then
			soop_error("method_proc",{ix,name,args},ERR_OBJ_ARGLENGTH)
		end if
		for x = 1 to length(name) do
			method_proc(ix,name[x],args[x])
		end for
	end if
	temp = soop_ResolvePoly(ix,name)
	if temp[1] = -1 then
		soop_error("method_proc",{ix,name,args},{ERR_OBJ_MEM404,temp[4],temp[2]})
	elsif temp[1] = 0 then
		soop_error("method_proc",{ix,name,args},ERR_OBJ_INVOBJ)
	else
		ix = temp[1]
		name = temp[2]
	end if
	m_ix = soop_FindPoly(ix,name,OBJECT_METHODS)
	if m_ix then
		oldThis = thisVar
		thisVar = ix
		call_proc(soop_MethodRID(ObjectClassIX[ix],m_ix),args)
		thisVar = oldThis
	else
		soop_error("method_proc",{ix,name,args},{ERR_OBJ_MET404,name,ix})
	end if
end procedure

-- /topic Object Manipulation
-- /func method_func( sObj Instance, eSeq MethodName, eSeq Arguments )
-- /ret Data from a Function
-- /desc Calls a Method on a Object Instance
--
-- This routine allows you to call a Method defined as an Euphoria
-- function.  It follows the same rules as call_func(), and the same
-- structure as method_proc().  When using multiple refrences, it
-- should be noted, that return values are returned in a sequence,
-- in the order that they are exected.
global function method_func(atom ix, sequence name, sequence args)
	sequence temp
	atom oldThis, m_ix
	object ret
	if ix = this then
		ix = thisVar
	end if
	if not soop_isString(name) then
		if length(name) = 1 then
			soop_error("method_func",{ix,name,args},ERR_STRING)
		end if
		if length(name) != length(args) then
			soop_error("method_func",{ix,name,args},ERR_OBJ_ARGLENGTH)
		end if
		for x = 1 to length(name) do
			method_proc(ix,name[x],args[x])
		end for
	end if
	temp = soop_ResolvePoly(ix,name)
	if temp[1] = -1 then
		soop_error("method_func",{ix,name,args},{ERR_OBJ_MEM404,temp[4],temp[2]})
	elsif temp[1] = 0 then
		soop_error("method_func",{ix,name,args},ERR_OBJ_INVOBJ)
	else
		ix = temp[1]
		name = temp[2]
	end if
	m_ix = soop_FindPoly(ix,name,OBJECT_METHODS)
	if m_ix then
		oldThis = thisVar
		thisVar = ix
		ret = call_func(soop_MethodRID(ObjectClassIX[ix],m_ix),args)
		thisVar = oldThis
		return ret
	else
		soop_error("method_func",{ix,name,args},{ERR_OBJ_MET404,name,ix})
	end if
end function

-- /topic Object Manipulation
-- /func method_id( sObj Instance, eSeq MethodName )
-- /ret A Unique Method ID
-- /desc Creates a Method ID on a Method declared on an Object Instance.
--
-- This function works in the same way that routine_id() does, however,
-- it is meant for the SOOP Library only, as it works only with
-- Methods declared on a Object.  What this does, is creates a Unique
-- ID from the Method Name, and Object Instance, in which to allow for
-- the patched in versions of call_func()/call_proc() to utilize.  The
-- reason for this, is to allow other libraries, such as win32lib, or
-- Allegro, to use the SOOP Enviroment, without any modification to
-- thoes libraries.  However, in order to allow SOOP's patched in versions
-- of call_func()/call_proc(), you will need to include soop.e before any
-- other library that you want to utilize this system.
--
-- See Demo3.exw for an example.
global function method_id(atom ix, sequence name)
	sequence temp
	atom mm_ix, m_ix
	if ix = this then
		ix = thisVar
	end if
	if not soop_isString(name) then
		if length(name) = 1 then
			soop_error("method_id",{ix,name},ERR_STRING)
		end if
		temp = {}
		for x = 1 to length(name) do
			temp &= method_id(ix,name)
		end for
		return temp
	end if
	temp = soop_ResolvePoly(ix,name)
	if temp[1] = -1 then
		soop_error("method_id",{ix,name},{ERR_OBJ_MEM404,temp[4],temp[2]})
	elsif temp[1] = 0 then
		soop_error("method_id",{ix,name},ERR_OBJ_INVOBJ)
	else
		ix = temp[1]
		name = temp[2]
	end if
	m_ix = soop_FindPoly(ix,name,OBJECT_METHODS)
	if m_ix then
		mm_ix = find(-1,MethodID)
		if not mm_ix then
			mm_ix = length(MethodID)+1
			MethodID &= -1
		end if
		MethodID[mm_ix] = {ix,soop_MethodRID(ObjectClassIX[ix],m_ix)}
		mm_ix -= (mm_ix*2)+10
		return mm_ix
	else
		soop_error("method_id",{ix,name},{ERR_OBJ_MET404,name,ix})
	end if
end function


-- For implementing method_id()'s into Euphoria's natural
-- call_func()/call_proc()
without warning
procedure euCallProc(integer id, sequence args)
	call_proc(id,args)
end procedure

function euCallFunc(integer id, sequence args)
	return call_func(id,args)
end function

global procedure call_proc(atom id, sequence args)
	atom oldThis
	if id < -10 then
		id -= (id*2)+10
		oldThis = thisVar
		thisVar = MethodID[id][1]
		euCallProc(MethodID[id][2],args)
		thisVar = oldThis
	else
		euCallProc(id,args)
	end if
end procedure

global function call_func(atom id, sequence args)
	object ret
	atom oldThis
	if id < -10 then
		id -= (id*2)+10
		oldThis = thisVar
		thisVar = MethodID[id][1]
		ret = euCallFunc(MethodID[id][2],args)
		thisVar = oldThis
		return ret
	else
		return euCallFunc(id,args)
	end if
end function

-- For implementing DLL compatable call_back()'s with
-- methods
function soop_CB(atom rid, atom nParams, atom lpParams)
	sequence params
	params = peek4u({lpParams,nParams})
	rid -= (rid*2)
	rid -= 10
	return call_func(rid,params)
end function

function soop_InitSTDCALL()
	atom stdcall
	stdcall = machine_func(16,27)
	poke(stdcall, {
	    #89,#E0,                --    0: mov eax, esp
	    #83,#C0,#04,            --    2: add eax, 4
	    #50,                    --    5: push eax
	    #68,#00,#00,#00,#00,    --    6: push dword nParams (7)
	    #68,#00,#00,#00,#00,    --    B: push dword rid (12)
	    #FF,#15,#00,#00,#00,#00,--   10: call near dword ptr [pfunc] (18)
	    #C3})                   --   16: ret
	                            --   17: bytes
	return stdcall
end function

procedure soop_StoreSTDCALL(atom stdcall, atom nParams, atom rid)
	poke4(stdcall+7,nParams)
	poke4(stdcall+12,rid)
	poke4(stdcall+18,stdcall+23)
	poke4(stdcall+23,machine_func(52,routine_id("soop_CB")))
end procedure

function soop_InitCDECL()
	atom cdecl
	cdecl = machine_func(16,30)
	poke(cdecl, {
	    #89,#E0,                --    0: mov eax, esp
	    #83,#C0,#04,            --    2: add eax, 4
	    #50,                    --    5: push eax
	    #68,#00,#00,#00,#00,    --    6: push dword nParams (7)
	    #68,#00,#00,#00,#00,    --    B: push dword rid (12)
	    #FF,#15,#00,#00,#00,#00,--   10: call near dword ptr [pfunc] (18)
	    #83,#EC,#08,            --   16: sub esp, 8
	    #C3})                   --   19: ret
	                            --   1A: bytes
	return cdecl
end function

procedure soop_StoreCDECL(atom cdecl, atom nParams, atom rid)
	poke4(cdecl + 7, nParams)
	poke4(cdecl + 12, rid)
	poke4(cdecl + 18,cdecl+26)
	poke4(cdecl + 26, machine_func(52,{'+',routine_id("soop_CB")}))
end procedure

-- /topic Object Manipulation
-- /func method_call_back( eObj RoutineID, eInt ParamCount )
-- /ret Returns a Method Call Back ID
-- /desc This creates a Method Call Back ID for use with External Libraries.
--
-- Like it's counterpart, call_back(), this creates a Memory address, with
-- assembly to be passed along to an External library, such as
-- Dynamic Link Library, or Shared Object, dependant apon the platform your
-- running on.  This allows External libraries to utilize the SOOP Enviroment,
-- without having to pass extra data to it.  The External Library never needs
-- to know which Object is being used by the SOOP library, in order to properly
-- interact with it, since the information is stored locally.  You must pass a
-- valid method_id() to method_call_back() in order for it to work, and you can
-- use '+' in a sequence pass for the ID, to declare a CDECL callback, instead of
-- a STDCALL.
global function method_call_back(object id, integer pCount)
	atom cbpoint
	if sequence(id) then
		id = id[2]
		if not (id < -10) then
			soop_error("method_call_back",{{'+',id},pCount},"Invalid method id given.")
		end if
		id -= (id*2)+10
		cbpoint = soop_InitCDECL()
		soop_StoreCDECL(cbpoint,pCount,id)
	else
		if not (id < -10) then
			soop_error("method_call_back",{id,pCount},"Invalid method id given.")
		end if
		id -= (id*2)+10
		cbpoint = soop_InitSTDCALL()
		soop_StoreSTDCALL(cbpoint,pCount,id)
	end if
	return cbpoint
end function

-- For Internal usage only, do not use without express knowledge

-- /topic Object Low Level Routines
-- /info
--  Low level routines that deal with the Object Enviroment
-- These functions allow for accessing the Object Enviroment,
-- and is mainly meant for use internally by the library.  Do
-- not use without express knowledge of the internal workings
-- of the SOOP Library.  They are provided without warranty.

-- /topic Object Low Level Routines
-- /func soop_OBJ_Num()
-- /ret Integer representing the number of Objects
-- /desc Returns the Number of objects in memory
--
-- This returns the Number of Objects in Memory, for
-- faster access to each object.  It was primarly
-- developed for the SOOP Debug Class, so that
-- it could debug the Object Enviroment.
global function soop_OBJ_Num()
	return length(ObjectClass)
end function

-- /topic Object Low Level Routines
-- /func soop_OBJ_EnumMembers( sObj Instance )
-- /ret Sequence of Members
-- /desc Returns a List of Members on an Object.
--
-- This returns the names of the Members on a Object, for
-- debugging information.  It returns the Current List of
-- members defined on an object.
global function soop_OBJ_EnumMembers(atom ix)
	return ObjectMembers[ix]
end function

-- /topic Object Low Level Routines
-- /func soop_OBJ_ClassIX( sObx Instance )
-- /ret Class Instance
-- /desc Returns the Class Instance associated with an Object.
--
-- This function returns the Class Instance assigned to an
-- object given.  It allows for faster retrival of certian
-- aspects, such a Member Types.
global function soop_OBJ_ClassIX(atom ix)
	return ObjectClassIX[ix]
end function

-- /topic Object Low Level Routines
-- /func soop_OBJ_MemberVal( sObj Instance, eAtom MemberInstance )
-- /ret Value of Member
-- /desc Returns the Value of a Member on an Object.
--
-- This function allows the faster retrival of Member data on
-- an object, mainly because there isn't as much overhead in
-- this function, as there is with oget().  It was mainly
-- used for the Debug System, so that as much speed
-- could be gained without having to resort to the oget() function.
global function soop_OBJ_MemberVal(atom ix, atom m_ix)
	return ObjectVTBL[ix][m_ix]
end function