-- Simplistic Object Oriented Programming (SOOP) API Library
-- Version 1.0 Beta
-- Created by: Mario Steele <eumario@trilake.net>
-- Website: http://enchantedblade.trilake.net/
--
-- Filename: soop_vars.e
-- Details: This system allows for Customizable Data Types, which allow
-- for easier definition of Classes.  Since SOOP's flexability allows for
-- any DataType to be placed in a class, it allows for easily adding types
-- of data, in which you can define the way data is being stored for later
-- use.  As long as a pointer can be retrieved, it'll be happy.

sequence TypeDefs,
		 TypeAssigns,
		 TypeReads,
		 TypeInit,
		 TypeFree,
		 TypeNoPoint

TypeDefs = {}
TypeAssigns = {}
TypeReads = {}
TypeInit = {}
TypeFree = {}
TypeNoPoint = {}
integer cType
cType = 0

-- /topic Data Type System
-- /info
--  The new way to do Types!
-- Okay, the reason why this is the new way to do Types, is
-- cause, not only can you check the data, to ensure that it'll
-- fit into the Variable/Member, it's also cause you can store
-- the data however you like!  This brings a new freedom to the
-- program, that wasn't previously available.  Some C-Structure
-- support should be possible through this method.  Future
-- versions may have the ability to support Structures better.

-- /topic Data Type System
-- /proc DataType( sfStr Name )
-- /desc This begins the definition of a Data Type
--
-- The Data Type system is setup just like the Class
-- definition system, in where you have a DataType() block
-- much like a for loop, if elsif end if, while loop, etc
-- etc.  You start a DataType block with this, by given
-- it a name, that can be later refrenced to when a error
-- occurs.  Like with Classes, Data Types cannot be
-- nested.  This is subject to change, to support Structures.
global procedure DataType(sequence name)
	if not soop_isString(name) then
		soop_error("DataType",{name},ERR_STRING)
	end if
	if cType then
		soop_error("DataType",{name},ERR_DT_INDEF)
	end if
	TypeDefs = append(TypeDefs,name)
	TypeAssigns = append(TypeAssigns,0)
	TypeReads = append(TypeReads,0)
	TypeInit = append(TypeInit,0)
	TypeFree = append(TypeFree,0)
	TypeNoPoint = append(TypeNoPoint,1)
	cType = length(TypeDefs)
end procedure

-- /topic Data Type System
-- /proc Assignment( eAtom RoutineID )
-- /desc Associates a Routine ID as an Assignment Call.
-- This routine associates an Euphoria function as a handler
-- for assigning data to the data type.  All type checking, and
-- actual storing of the data is done with the routine id given.
global procedure Assignment(atom rid)
	if not cType then
		soop_error("Assignment",{rid},ERR_DT_NINDEF)
	end if
	TypeAssigns[cType] = rid
end procedure

-- /topic Data Type System
-- /proc Retrival( eAtom RoutineID )
-- /desc Associates a Routine ID as an Retrival Call.
-- This routine associates an Euphoria functon as a handler
-- for retriving data from the data type.
global procedure Retrival(atom rid)
	if not cType then
		soop_error("Retrival",{rid},ERR_DT_NINDEF)
	end if
	TypeReads[cType] = rid
end procedure

-- /topic Data Type System
-- /proc VarInit( eAtom RoutineID )
-- /desc Associates a Routine ID as an Initalization Call.
-- This routine associates an Euphoria function as a handler
-- to initalize the data type, for future usage by the system.
global procedure VarInit(atom rid)
	if not cType then
		soop_error("VarInit",{rid},ERR_DT_NINDEF)
	end if
	TypeInit[cType] = rid
end procedure

-- /topic Data Type System
-- /proc VarFree( eAtom RoutineID )
-- /desc Associates a Routine ID as an Freeing Call.
-- Much like with Memory, you have to free it to return it to
-- the system for future use, this allows you to write a routine
-- to associate for a data type, to do any cleanup that may
-- or may not be needed.
global procedure VarFree(atom rid)
	if not cType then
		soop_error("VarFree",{rid},ERR_DT_NINDEF)
	end if
	TypeFree[cType] = rid
end procedure

-- /topic Data Type System
-- /proc DataTypeID()
-- /desc Returns the Data Type ID for use.
-- Since Most declorations in this are procedures, you
-- can use DataTypeID() to retrive the ID for use with
-- the DataType system.  If you do not have the ID, you
-- cannot use it.  This should generally be declared as
-- a global constant, so that any file in your program
-- can use it, especially if the Data Type is declared in
-- a seperate library.
global function DataTypeID()
	if not cType then
		soop_error("DataType",{},ERR_DT_NINDEF)
	end if
	return cType
end function

-- /topic Data Type System
-- /proc NoPointer()
-- /desc Dis-allows for Pointer Retrival.
-- This is mainly here, so that you can prevent retrival of
-- the Pointer to the data. This may and may not be useful,
-- but as demonstrated with the STRING data type, it can be
-- useful, since the memory address to the string can be used
-- by other things, such a External Libraries.
global procedure NoPointer()
	if not cType then
		soop_error("DataType",{},ERR_DT_NINDEF)
	end if
	TypeNoPoint[cType] = 0
end procedure

-- /topic Data Type System
-- /proc EndDataType()
-- /desc Ends the DataType Definition
-- This pretty much closes the Data Type Definition block, and 
-- should be called when you are done with the data type definition.
global procedure EndDataType()
	cType = 0
end procedure

-- /topic Data Type Low Level
-- /info
--  The Core of how Data Types Work
-- These are the functions that allow Data Types to be used in Euphoria,
-- and in SOOP in general.  Your free to use the Data Type system in
-- another project, should you choose to.  But please leave credit where
-- it is due.  These functions are provided to help assist the SOOP library
-- in actually storing and retriving the data.

-- /topic Data Type Low Level
-- /func DataType_Store( dt DataTypeID, eObj Original, eObj New)
-- /ret Actual data, or pointer to said data.
-- /desc Calls the handler for the Assignment of data.
--
-- This function actually allows the handlers associated with the
-- data type to assign data.  All type checking should be done at
-- this time, to ensure that the data is allowed.  If everything
-- is successful, the pointer to the data or the actual data
-- should be returned.  The pointer should be meaningful to your
-- functions though.
global function DataType_Store(atom vType, object orig, object new)
	if vType > 0 and vType < length(TypeDefs)+1 then
		return call_func(TypeAssigns[vType],{orig,new})
	else
		soop_error("[INTERNAL:DataType] Assignment",{},ERR_DT_INVTYPE)
	end if
end function

-- /topic Data Type Low Level
-- /func DataType_Retrive( dt DataTypeID, eObj Pointer )
-- /ret Actual data stored
-- /desc This retrives the actual data stored in a Data Type
--
-- This function actually allows the handlers associated with the
-- data type to retrive the data.  There shouldn't be any type
-- checking done during this time, and should only be done in
-- the assignment.
global function DataType_Retrieve(atom vType, object dat)
	if vType > 0 and vType < length(TypeDefs)+1 then
		return call_func(TypeReads[vType],{dat})
	else
		soop_error("[INTERNAL:DataType] Retrival",{},ERR_DT_INVTYPE)
	end if
end function

-- /topic Data Type Low Level
-- /func DataType_AllowPointer( dt DataTypeID )
-- /ret Boolean True or False
-- /desc Check to see if a DataType allows a Pointer to be returned.
--
-- This function is a support function, to ensure that the data type
-- given, allows the pointer to be returned, instead of the actual
-- data.
global function DataType_AllowPointer(atom vType)
	if vType > 0 and vType < length(TypeDefs)+1 then
		return TypeNoPoint[vType]
	else
		soop_error("[INTERNAL:DataType] AllowPointer",{},ERR_DT_INVTYPE)
	end if
end function

-- /topic Data Type Low Level
-- /func DataType_InitVar( dt DataTypeID )
-- /ret Inital pointer for the data
-- /desc Initalizes the Data Type for use.
--
-- This allows the associated routines, to initalize the variable space
-- for a data type, so that there are no errors, when the user of the
-- data type, attempts to assign data to it.
global function DataType_InitVar(atom vType)
	if vType > 0 and vType < length(TypeDefs)+1 then
		if TypeInit[vType] = 0 then
			return 0
		else
			return call_func(TypeInit[vType],{})
		end if
	else
		soop_error("[INTERNAL:DataType] InitVar",{},ERR_DT_INVTYPE)
	end if
end function

-- /topic Data Type Low Level
-- /proc DataType_FreeVar( dt DataTypeID )
-- /desc Releases the resources for the DataType variable space.
--
-- This aloows the associated routines, to release the variable space
-- for a data type, to be executed.
global procedure DataType_FreeVar(atom vType, object val)
	if vType > 0 and vType < length(TypeDefs)+1 then
		if TypeFree[vType] != 0 then
			call_proc(TypeFree[vType],{val})
		end if
	else
		soop_error("[INTERNAL:DataType] FreeVar",{},ERR_DT_INVTYPE)
	end if
end procedure

-- /topic Data Type Low Level
-- /func DataType_GetTypeName( dt DataTypeID )
-- /ret String Representing the DataType Name
-- /desc Returns the Data Type's name, as defined.
--
-- This function is mainly here to allow DataType names to be used
-- when Debug Information is being written out by the SOOP Debug Class.
global function DataType_GetTypeName(atom vType)
	if vType > 0 and vType < length(TypeDefs)+1 then
		return TypeDefs[vType]
	else
		soop_error("[INTERNAL:DataType] GetTypeName",{},ERR_DT_INVTYPE)
	end if
end function


-- Default Euphoria Types

-- /topic SOOP Data Types
-- /info
--  Data Types implemented for the SOOP Library.
--
-- These data types are meant for use in the SOOP Library, these
-- types include the 4 basic types that Euphoria supports by
-- default, as well as a special type, to show how to implement
-- the pointer system with SOOP, and the Data Type System.

-- /topic SOOP Data Types
-- /const INTEGER
-- /desc Euphoria's Integer Data Type
--
-- The INTEGER data type allows for Integer data to be stored
-- in a SOOP Member.  It works exactly the same as a normal
-- Euphoria integer type.
DataType("INTEGER")
	function INTEGER_Assign(integer orig, integer new)
		orig = orig
		return new
	end function
	
	function INTEGER_Retrieve(integer orig)
		return orig
	end function
	
	function INTEGER_Init()
		return 0
	end function
	
	global constant INTEGER = DataTypeID()
	Assignment(routine_id("INTEGER_Assign"))
	Retrival(routine_id("INTEGER_Retrieve"))
	VarInit(routine_id("INTEGER_Init"))
	NoPointer()
EndDataType()

-- /topic SOOP Data Types
-- /const ATOM
-- /desc Euphoria's Atomic Data Type
--
-- The ATOM data type allows for Atomic data to be stored
-- in a SOOP Member.  It works exactly the same as a normal
-- Euphoria atom type.
DataType("ATOM")
	function ATOM_Assign(atom orig, atom new)
		orig = orig
		return new
	end function
	
	function ATOM_Retrieve(atom orig)
		return orig
	end function
	
	function ATOM_Init()
		return 0
	end function
	
	global constant ATOM = DataTypeID()
	Assignment(routine_id("ATOM_Assign"))
	Retrival(routine_id("ATOM_Retrieve"))
	VarInit(routine_id("ATOM_Init"))
	NoPointer()
EndDataType()

-- /topic SOOP Data Types
-- /const SEQUENCE
-- /desc Euphoria's Sequence Data Type
--
-- The SEQUENCE data type allows for Sequence data to be stored
-- in a SOOP Member.  It works exactly the same as a normal
-- Euphoria sequence type.
DataType("SEQUENCE")
	function SEQUENCE_Assign(sequence orig, sequence new)
		orig = orig
		return new
	end function
	
	function SEQUENCE_Retrieve(sequence orig)
		return orig
	end function
	
	function SEQUENCE_Init()
		return {}
	end function
	
	global constant SEQUENCE = DataTypeID()
	Assignment(routine_id("SEQUENCE_Assign"))
	Retrival(routine_id("SEQUENCE_Retrieve"))
	VarInit(routine_id("SEQUENCE_Init"))
	NoPointer()
EndDataType()

-- /topic SOOP Data Type
-- /const OBJECT
-- /desc Euphoria's Object Data Type
--
-- The OBJECT data type allows for any of the 3 main data types
-- in Euphoria to be assigned to, just like with it's Euphoria
-- counterpart.
DataType("OBJECT")
	function OBJECT_Assign(object orig, object new)
		orig = orig
		return new
	end function
	
	function OBJECT_Retrieve(object orig)
		return orig
	end function
	
	function OBJECT_Init()
		return 0
	end function
	
	global constant OBJECT = DataTypeID()
	Assignment(routine_id("OBJECT_Assign"))
	Retrival(routine_id("OBJECT_Retrieve"))
	VarInit(routine_id("OBJECT_Init"))
	NoPointer()
EndDataType()

-- Custom Data Type

-- /topic SOOP Data Types
-- /const STRING
-- /desc A C-Style String Data Type
--
-- This is the custom Data Type, that shows how pointers can be
-- useful.  It allows for a Normal C-Style string to be stored
-- into memory, and therefore, allows users to store and retrive
-- strings into it, or retrive the Memory Address of the string
-- to pass along to C/C++ Functions that require it.
DataType("STRING")
	function STRING_Assign(atom lpszOrig, sequence lpszNew)
		atom lpsz
		if lpszOrig != -1 then
			machine_proc(17,lpszOrig)
		end if
		lpsz = machine_func(16,length(lpszNew)+1)
		poke(lpsz,lpszNew & 0)
		return lpsz
	end function
	
	function STRING_Retrieve(atom lpsz)
		sequence dat
		integer o
		if lpsz <= 0 then
			return -1
		end if
		dat = {}
		o = 0
		while peek(lpsz+o) do
			dat &= peek(lpsz+o)
			o += 1
		end while
		return dat
	end function
	
	function STRING_Init()
		return -1
	end function
	
	procedure STRING_Free(atom lpsz)
		if lpsz != -1 then
			machine_proc(17,lpsz)
		end if
	end procedure
	
	global constant STRING = DataTypeID()
	Assignment(routine_id("STRING_Assign"))
	Retrival(routine_id("STRING_Retrieve"))
	VarInit(routine_id("STRING_Init"))
	VarFree(routine_id("STRING_Free"))
EndDataType()